/* displayserver.c - simple server for TCP/IP sockets */

#ifdef MEMWATCH
#include "memwatch.h"
#endif
#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>
#include <sched.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/vfs.h>

struct dsprocess {
	int pid;
	int type;
	int totsz;
	int compl;
};

struct logentry {
	int pid;
	int remote;
	time_t time;
	char *mesg;
	struct logentry *next;
	struct logentry *prev;
};

struct displayserv {
	struct dsprocess *ps;
	struct logentry *le;
	struct logentry *lle;
	struct logentry *fle;
	int logcount;
};

struct displayservconn {
	struct displayserv *ds;
	int conn;
	char *vhost;
	FILE *fp;
};

#define PROC_SUPER_MAGIC 0x9fa0
#define MYBUFSIZ	1024
#define MAXLOG		100
#define MINSLEEP	50000
#define MAXSLEEP	10000000
#define DEFSLEEP	100000
#define LISTEN_SOCKET	80
#define EMPEG_INDEX	1
#define EMPEG_JSCRIPT	2
#define EMPEG_HTML	3
#define EMPEG_CLASS	4
#define EMPEG_FASCIA	5
#define EMPEG_STREAM	6
#define EMPEG_MP3PLIST	7
#define EMPEG_MP3FILE	8
#define EMPEG_M3U	9
#define EMPEG_SYSINFO	10
#define EMPEG_LOG	11
#define CLASSFILE	"empegApplet.class"
#define FASCIAFILE	"fascia.jpg"

int b64_pton(src, target, targsize)
/*
Taken from openldap - base64.c
*/
	char const *src;
	u_char *target;
	size_t targsize;
{
	static const char Base64[] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	static const char Pad64 = '=';
	int tarindex, state, ch;
	char *pos;

	state = 0;
	tarindex = 0;


	while ((ch = *src++) != '\0') {
		if (isspace(ch))	/* Skip whitespace anywhere. */
			continue;

		if (ch == Pad64)
			break;

		pos = strchr(Base64, ch);
		if (pos == 0) 		/* A non-base64 character. */
			return (-1);

		switch (state) {
		case 0:
			if (target) {
				if (tarindex >= targsize)
					return (-1);
				target[tarindex] = (pos - Base64) << 2;
			}
			state = 1;
			break;
		case 1:
			if (target) {
				if (tarindex + 1 >= targsize)
					return (-1);
				target[tarindex]   |=  (pos - Base64) >> 4;
				target[tarindex+1]  = ((pos - Base64) & 0x0f)
							<< 4 ;
			}
			tarindex++;
			state = 2;
			break;
		case 2:
			if (target) {
				if (tarindex + 1 >= targsize)
					return (-1);
				target[tarindex]   |=  (pos - Base64) >> 2;
				target[tarindex+1]  = ((pos - Base64) & 0x03)
							<< 6;
			}
			tarindex++;
			state = 3;
			break;
		case 3:
			if (target) {
				if (tarindex >= targsize)
					return (-1);
				target[tarindex] |= (pos - Base64);
			}
			tarindex++;
			state = 0;
			break;
		default:
			abort();
		}
	}

	/*
	 * We are done decoding Base-64 chars.  Let's see if we ended
	 * on a byte boundary, and/or with erroneous trailing characters.
	 */

	if (ch == Pad64) {		/* We got a pad char. */
		ch = *src++;		/* Skip it, get next. */
		switch (state) {
		case 0:		/* Invalid = in first position */
		case 1:		/* Invalid = in second position */
			return (-1);

		case 2:		/* Valid, means one byte of info */
			/* Skip any number of spaces. */
			for (NULL; ch != '\0'; ch = *src++)
				if (!isspace(ch))
					break;
			/* Make sure there is another trailing = sign. */
			if (ch != Pad64)

			ch = *src++;		/* Skip the = */
			/* Fall through to "single trailing =" case. */
			/* FALLTHROUGH */

		case 3:		/* Valid, means two bytes of info */
			/*
			 * We know this char is an =.  Is there anything but
			 * whitespace after it?
			 */
			for (NULL; ch != '\0'; ch = *src++)
				if (!isspace(ch))
					return (-1);

			/*
			 * Now make sure for cases 2 and 3 that the "extra"
			 * bits that slopped past the last full byte were
			 * zeros.  If we don't check them, they become a
			 * subliminal channel.
			 */
			if (target && target[tarindex] != 0)
				return (-1);
		}
	} else {
		/*
		 * We ended by seeing the end of the string.  Make sure we
		 * have no partial bytes lying around.
		 */
		if (state != 0)
			return (-1);
	}

	return (tarindex);
}

int readconfig() {
/* read config file */
	FILE *fp;
	char s[256];
	
	fp=fopen("config","r");
	if ( (fp=fopen("config", "r")) == NULL ) {
		fprintf (stderr, "ttt\n");
	} else {
        	while (fgets(s, sizeof(s),fp) != NULL ) {
			fprintf(stderr,"%s",s);
		}
	}
}

int fputs_url_encode(char *src, FILE *fp) {
/*
I found some trouble using spaces in URL's in m3u files for some players.
Maybe add some other encodings later
*/

        int ch,i=0;

        while ((ch = *src++) != '\0') {
                if ( isspace(ch) && (ch != 10) )
			fputs("%20",fp);
                else 
			fputc(ch,fp);
		i++;
        }
        return 0;
}

int logmsg(struct displayserv *ds, char *mesg) {
/*
Under construction
*/
	struct logentry *le;

	fprintf( stderr,"log: %s\n",mesg);
	le = (struct logentry *)malloc(sizeof(struct logentry));
	le->time=time();
	le->mesg=strdup(mesg);
	fprintf( stderr,"log1\n");
	if ( (ds->lle != NULL) && (ds->lle->next != NULL) ) {
		le->next=ds->lle->next->next;
		le->prev=ds->lle;
		le->prev->next=le;
		le->next->prev=le;
		free(ds->lle->next->mesg);
		free(ds->lle->next);
		ds->lle=le;
		ds->fle=le->next;
	} else {
	fprintf( stderr,"log2\n");
		if ( ds->fle == NULL ) {
			ds->fle = le;
			ds->lle = le;
		}
	fprintf( stderr,"log3\n");
		if ( ds->logcount > MAXLOG ) {
			le->next=ds->fle;
		} else
			le->next=NULL;
	fprintf( stderr,"log4\n");
		ds->lle->next=le;
		le->prev=ds->lle;
		ds->lle=le;
	}
	ds->logcount++;
	fprintf( stderr,"fle: %d\n",ds->fle);
	fprintf( stderr,"lle: %d\n",ds->lle);
	fprintf( stderr,"lle: %d\n",ds->logcount);
};

int serve_stream(struct displayservconn *dsconn,  int sl) {
/*
serve display data to applet
*/
	int childpid=fork();

	if ( childpid > 0 ) {

		int disfd;
		int i,cont=1;
		char buf[2048];
		char nbuf[1024];
		caddr_t dmap;

		disfd = open("/dev/display", O_RDWR);
		if (disfd == -1) {
			fprintf( stderr,"Error opening /dev/display\n");
			return 1;
		}

		dmap=mmap(0, 2048, PROT_READ|PROT_WRITE, MAP_SHARED, disfd, 0);
		if (dmap == (caddr_t) -1) {
			fprintf( stderr, "Error mmapping /dev/display\n");
			return 4;
		}

		send_httpheader(dsconn->fp,0,"application/octet-stream");
		while ( cont ) {
			memcpy(&buf,dmap,2048);
			for (i=0;i<1024;i++) {
				nbuf[i]=(
			(  buf[i*2]           & 3 ) |
			(  buf[i*2] >> 2      & 12 ) |
			( (buf[i*2+1] << 4  ) & 48 ) |
			( (buf[i*2+1] << 2  ) & 192 )
			);
			}
			if(write(dsconn->conn,nbuf,sizeof(nbuf))!=sizeof(nbuf))
				cont=0;
			usleep(sl);
		}
		munmap(dmap, 2048);
		close(disfd);
	} else {

		int irfd;
		ssize_t rts;
		char buf[8];
		struct empeg_ir_write {
			int type;
			long ircode;
		} *irdata=(struct empeg_ir_write *)
			malloc( sizeof(struct empeg_ir_write));

		while (strlen(fgets( buf,sizeof(buf), dsconn->fp)) > 2 ) {
			irdata->type=atoi(strtok(buf," "));	
			irdata->ircode=atoi(strtok(NULL,"\n"));	
			irfd = open("/dev/ir", O_WRONLY|O_CREAT|O_TRUNC);
			rts = write(irfd, irdata, sizeof(*irdata));
			if (rts == -1 )
				perror("Writing to IR device.");
			close(irfd);
#ifdef DEBUG
			fprintf( stderr,"type: %d\n",irdata->type);
			fprintf( stderr,"ircode: %d\n",irdata->ircode);
#endif
		}
	free(irdata);
	}
	return 0;
}
	
int send_jscript(FILE *fp) {
	
	send_httpheader(fp,0,"text/html");
	fputs("<HTML><SCRIPT LANGUAGE='JavaScript'>",fp);
	fputs("if (window.empegwindow) empegwindow.close();",fp);
	fputs("options='toolbar=0,status=1,menubar=0,scrollbars=0,",fp);
        fputs("resizable=no,width=330,height=125';",fp);
	fputs("empegwindow=window.open(",fp);
	fputs("'empeg.html','empegwindow', options);",fp);
	fputs("empegwindow.focus();history.back();</SCRIPT></HTML>\n",fp);
	return 0;
}

int send_index(struct displayservconn *dsconn) {
	
	FILE *fp=dsconn->fp;
	send_httpheader(fp,0,"text/html");
	fputs("<HTML><TITLE>Empeg WebUtils</TITLE>",fp);
	fputs("<BODY BGCOLOR=\"#000000\" ",fp);
	fputs("LINK=\"#ffc000\" VLINK=\"#ffc000\" ALINK=\"red\">",fp);
	fputs("<TABLE BORDER=0 CELLPADDING=0 ALIGN=CENTER ",fp);
	fputs("VALIGN=CENTER CELLSPACING=0>",fp);
        fputs("<TR><TD ALIGN=CENTER VALIGN=CENTER>",fp);
	fputs("<H1><font color=\"#ffcc00\">Empeg WebUtils</font></H1>",fp);
	fputs("<LI><A HREF=\"/jscript.html\">empegApplet in new window</A>",fp);
	fputs("<LI><A HREF=\"/empeg.html\">empegApplet in current window</A>",fp);
	fputs("<LI><A HREF=\"/mp3list.html\">MP3 list</A>",fp);
	fputs("<LI><A HREF=\"/sysinfo.html\">System Information</A>",fp);
	fputs("</TD></TR></TABLE></BODY></HTML>",fp);
	return 0;
}

int send_html(FILE *fp) {
	
	send_httpheader(fp,0,"text/html");
	fputs("<HTML><TITLE>Empeg Applet</TITLE>",fp);
	fputs("<BODY BGCOLOR=\"#000000\" ",fp);
	fputs("LINK=\"#ffc000\" VLINK=\"#ffc000\" ALINK=\"red\">",fp);
	fputs("<TABLE BORDER=0 CELLPADDING=0 ALIGN=CENTER ",fp);
	fputs("VALIGN=CENTER CELLSPACING=0>",fp);
        fputs("<TR><TD ALIGN=CENTER VALIGN=CENTER>",fp);
	fputs("<APPLET CODE=\"empegApplet.class\" HEIGHT=93 WIDTH=296>",fp);
	fputs("<PARAM NAME=sleeptime VALUE=\"10000\">",fp);
	fputs("<PARAM NAME=color0 VALUE=\"#000000\">",fp);
	fputs("<PARAM NAME=color1 VALUE=\"#007700\">",fp);
	fputs("<PARAM NAME=color2 VALUE=\"#00AA00\">",fp);
	fputs("<PARAM NAME=color3 VALUE=\"#00FF00\">",fp);
	fputs("</APPLET></TD></TR></TABLE></BODY></HTML>",fp);
	return 0;
}

int send_mp3fids(struct displayservconn *dsconn, char *mp3fid, int m3u) {
/*
Routine to generate playlist listing or generate m3u file
*/
	FILE *fp=dsconn->fp;
	DIR *dir;
	FILE *fidfp, *fidfp2;
	struct dirent *d;
	int i,drive1,m3ulink=0;
	char fid[4];
	char lfilename[8];
	char *sfilename;
	char *c,*type,*album,*artist,*title;
	char s[256];
	char t[1024];
	char e[1]="\0";

	if ( (chdir("/drive1/fids")) != -1 ) {
		drive1=1;
		if ( (fidfp=fopen(mp3fid, "r")) == NULL ) {
			chdir("/drive0/fids");
			fidfp=fopen(mp3fid, "r");
		}
	} else {
		chdir("/drive0/fids");
		fidfp=fopen(mp3fid, "r");
	}
	if ( fidfp != NULL ) {
        	while (fread(fid, sizeof(fid),1,fidfp) != 0 ) {
			sprintf(lfilename, "%02x%02x%02x%02x",
					fid[3],fid[2],fid[1],fid[0]);
			sfilename = (char *) &lfilename;
			i=0;
			while ( lfilename[i] == 48 ) {
				sfilename++;
				i++;
				}
			sfilename[strlen(sfilename)-1]='1';
			if ( drive1 ) {
				chdir("/drive1/fids");
				if ( (fidfp2=fopen(sfilename, "r")) == NULL ) {
					chdir("/drive0/fids");
					fidfp2=fopen(sfilename, "r");
				}
			} else
				fidfp2=fopen(sfilename, "r");
			if ( fidfp2 != NULL ) {
				type=album=title=artist=NULL;
        			while (fgets(s, sizeof(s),fidfp2) != NULL ) {
					if ( (strstr(s,"type=")!=NULL) ||
						(strstr(s,"Type=")!=NULL) )
						type=(char *)strdup(
							strtok(s+5,"\n"));
					if ( (strstr(s,"source=")!=NULL) ||
						(strstr(s,"Source=")!=NULL) )
						album=(char *)strdup(
							strtok(s+7,"\n"));
					if ( (strstr(s,"title=")!=NULL) ||
						(strstr(s,"Title=")!=NULL) )
						title=(char *)strdup(
							strtok(s+6,"\n"));
					if ( (strstr(s,"artist=")!=NULL) ||
						(strstr(s,"artist=")!=NULL) )
						artist=(char *)strdup(
							strtok(s+7,"\n"));
				}
#ifdef DEBUG
				fprintf(stderr,"type: %s\n",type);
				fprintf(stderr,"source: %s\n",album);
				fprintf(stderr,"title: %s\n",title);
				fprintf(stderr,"artist: %s\n",artist);
#endif
				if ( strstr(type,"tune") != NULL ) {
				if ( m3u ) {
					sprintf(t,
					    "http://%s/%s-%s-%s.mp3?file=%s\n",
					    dsconn->vhost,
						artist,album,title,sfilename);
					fputs_url_encode(t,fp);
				} else {
					fprintf(fp,
					"<A HREF=\"/%s-%s-%s.mp3?file=%s\">",
						artist,album,title,sfilename);
					fprintf(fp,"%s - %s - %s</A><BR>\n",
						artist,album,title);
					m3ulink=1;
				}
				}
				if ( (strstr(type,"playlist")!=NULL) && ! m3u) {
				fprintf(fp,"<A HREF=\"/%s?playlist=%s\">",
					title,sfilename);
				fprintf(fp,"PLAYLIST: %s</A><BR>\n", title);
				}
				if (type != NULL) free(type);
				if (artist != NULL) free(artist);
				if (album != NULL) free(album);
				if (title != NULL) free(title);
				fclose(fidfp2);
#ifdef DEBUG
				fprintf(stderr,"filename: %s\n",sfilename);
#endif
			}
		}
		fclose(fidfp);
		if ( m3ulink ) {
			fprintf(fp,"<A HREF=\"/emp.m3u?playlist=%s\">",mp3fid);
			fprintf(fp,"M3U</A><BR>\n");
		}
#ifdef DEBUG
		fprintf(stderr,"\n");
#endif
	}
	return 0;
}

int send_mp3list(struct displayservconn *dsconn, char *mp3fid) {
/*
playlist headers & footers
*/	
	FILE *fp=dsconn->fp;
	send_httpheader(fp,0,"text/html");
	fputs("<HTML><TITLE>Empeg MP3list</TITLE>",fp);
	fputs("<BODY BGCOLOR=\"#000000\" ",fp);
	fputs("LINK=\"#ffc000\" VLINK=\"#ffc000\" ALINK=\"red\">",fp);
	fputs("<H1><CENTER>",fp);
	fputs("<font color=\"#ffcc00\">Empeg MP3list</font>",fp);
	fputs("</CENTER></H1>",fp);
	send_mp3fids(dsconn,mp3fid,0);
	fputs("</BODY></HTML>",fp);
	return 0;
}

int send_m3u(struct displayservconn *dsconn, char *mp3fid) {
/*
serve m3ufile
*/	
#ifdef DEBUG
	fprintf( stderr,"m3u : %s\n",mp3fid);
#endif
	send_httpheader(dsconn->fp,0,"audio/x-mpegurl");
	send_mp3fids(dsconn, mp3fid, 1);
	return 0;
}

int send_httpheader(FILE *fp, int size, char *ct) {

	fputs("HTTP/1.1 200 OK\n",fp);
	fputs("Server: EmpegCar DisplayServer/1.1 (Linux)\n",fp);
	if ( size > 0 )
		fprintf(fp,"Content-Length: %d\n",size);
	fputs("Connection: Keep-Alive\n", fp);
	fprintf(fp,"Content-Type: %s\n\n", ct);
	return 0;
}

int serve_file(struct displayservconn *dsconn, char *filename) {
	
	FILE *fp=dsconn->fp;
	int bufSize,fd;
	char buf[MYBUFSIZ];
	struct stat *fdstat;

	fd = open(filename, O_RDONLY);
	if (fd == -1)
		exit(1);
	fdstat=(struct stat *) malloc( sizeof(struct stat) );
	if ( fstat(fd,fdstat) == -1 ) {
		free(fdstat);
		close(fd);
		exit(2);
	}
	if ( strstr(filename,".htm")!=NULL)
		send_httpheader(fp,fdstat->st_size,"text/html");
	else {
		if ( strstr(filename,"/fids/")!=NULL) 
			send_httpheader(fp,fdstat->st_size,"audio/x-mpeg");
		else
			send_httpheader(fp,fdstat->st_size,
						"application/octet-stream");
		}
	fflush(fp);
	while ( (bufSize=read(fd, buf, MYBUFSIZ) ) == 
				MYBUFSIZ ) {
		write(dsconn->conn, buf, bufSize);
		};
	if ( bufSize > 0 )
		write(dsconn->conn, buf, bufSize);
	free (fdstat);
	close(fd);
	return 0;
}

int send_log(struct displayservconn *dsconn) {
/*
under construction
*/
	struct logentry *le;
	struct displayserv *ds;
	FILE *fp=dsconn->fp;

	ds=dsconn->ds;
	fprintf( stderr,"fle: %d\n",ds->fle);
	fprintf( stderr,"lle: %d\n",ds->lle);
	fprintf( stderr,"lle: %d\n",ds->logcount);
	send_httpheader(fp,0,"text/html");
	fputs("<HTML><TITLE>Empeg Sysinfo</TITLE>",fp);
	fputs("<BODY TEXT=\"#EEEEEE\" BGCOLOR=\"#000000\" ",fp);
	fputs("LINK=\"#ffc000\" VLINK=\"#ffc000\" ALINK=\"red\">",fp);
	fputs("<CENTER><H1>",fp);
	fputs("<font color=\"#ffcc00\">Empeg Log</H1>",fp);
	if ( ds->fle == NULL ) {
		fputs("No entries found</BODY></HTML>",fp);
		return 1;
	}
	le=ds->fle;
	while (1) {
		fprintf(fp,"%s %s\n",ctime(le->time),le->mesg);
		if ( le == ds->lle )
			return 0;
		le=le->next;
	}
}

int get_psentry(FILE *fp, int pid) {
/*
get info from pid
*/
	char buffer[12];
	FILE *statfp;
	char state[1], comm[256];
	int i, rpid, ppid, pgrp, session, tty, tpgid, utime, stime, cutime,
		cstime, counter, priority, starttime, signal, blocked,
		sigignore, sigcatch;
	unsigned int flags, minflt, cminflt, majflt, cmajflt,
		timeout, itrealvalue, vsize, rss, rlim, startcode,
		endcode, startstack, kstkesp, kstkeip, wchan;

	sprintf(buffer, "%d/stat", pid);
	statfp=fopen(buffer, "r");
	sprintf(buffer, "%d/fd/2", pid);
	if ( (i=readlink(buffer, buffer, 12)) > 0 ) buffer[i]='\0';
	else strcpy(buffer,"/dev/?");
	if ( strncmp("/dev/tty",buffer,8) != 0 ) strcpy(buffer,"/dev/?");

	if (statfp == NULL)
		return 2;
	fscanf(statfp,"%d %s %c %d %d %d %d %d %u %u %u %u %u %d %d %d %d %d %d %u %u %d %u %u %u %u %u %u %u %u %d %d %d %d %u",
		&rpid, comm, state, &ppid, &pgrp, &session, &tty,
		&tpgid, &flags, &minflt, &cminflt, &majflt, &cmajflt,
		&utime, &stime, &cutime, &cstime, &counter, &priority,
		&timeout, &itrealvalue, &starttime, &vsize, &rss,
		&rlim, &startcode, &endcode, &startstack, &kstkesp,
		&kstkeip, &signal, &blocked, &sigignore, &sigcatch,
		&wchan);
	fclose(statfp);
	fprintf(fp,"<TR><TD>%s</TD><TD>%.1s</TD><TD>%d</TD><TD>%d</TD>",
		comm, state, rpid, ppid );
	fprintf(fp,"<TD>%d</TD><TD>%d</TD><TD>%d</TD><TD>%s</TD>",
		vsize/1024, rss*4, priority, &buffer[5] );
	fprintf(fp,"<TD>%d</TD><TD>%d</TD><TD>%d</TD><TD>%d</TD>",
		minflt, majflt, cminflt,cmajflt );
	fprintf(fp,"<TD>%d</TD><TD>%d</TD><TD>%d</TD><TD>%d</TD>",
		utime, stime, cutime, cstime );
	fprintf(fp,"</TR>\n");
}

int send_sysinfo(struct displayservconn * dsconn) {
/*
send sysinfo
*/
	FILE *fp=dsconn->fp;
	FILE *statfp;
	struct statfs sb;
	char buffer[4096];
	int fd,len;
	char *p;
	DIR *dir;
	struct dirent *ent;
	int pid, i;


	send_httpheader(fp,0,"text/html");
	if (statfs("/proc", &sb) < 0 || sb.f_type != PROC_SUPER_MAGIC)
	{
		fputs("<HTML><TITLE>Error</TITLE>",fp);
		fputs("<BODY TEXT=\"#EEEEEE\" BGCOLOR=\"#000000\" ",fp);
		fputs("<H1><CENTER>",fp);
		fputs("<font color=\"#ffcc00\">",fp);
		fputs("proc filesystem not mounted on /proc",fp);
		fputs("</CENTER></H1></FONT>",fp);
		fputs("</BODY></HTML>",fp);
		fprintf(stderr, "proc filesystem not mounted on /proc\n");
		return -1;
        }
	chdir("/proc");
	fputs("<HTML><TITLE>Empeg Sysinfo</TITLE>",fp);
	fputs("<BODY TEXT=\"#EEEEEE\" BGCOLOR=\"#000000\" ",fp);
	fputs("LINK=\"#ffc000\" VLINK=\"#ffc000\" ALINK=\"red\">",fp);
	fputs("<H1><CENTER>",fp);
	fputs("<font color=\"#ffcc00\">Empeg Sysinfo",fp);

	fputs("<P><TABLE BORDER=1>",fp);
	statfp = fopen("empeg_therm", "r");
	if (fgets(buffer, sizeof(buffer),statfp) != NULL)
	fprintf(fp,
	     "<TR><TD>Empeg Temperature</TD><TD>%s degrees Celcius</TD></TR>\n",
		buffer);
	fputs("</TABLE>",fp);

	fputs("<P><TABLE BORDER=1>",fp);
	fputs("<TR><TH COLSPAN=2>Empeg ID</TH></TR>",fp);
	statfp = fopen("empeg_id", "r");
	while(fgets(buffer, sizeof(buffer),statfp) != NULL) {
		fprintf(fp, "<TR><TD>%s</TD><TD>%s</TD></TR>\n",
		strtok(buffer, ":"),strtok(NULL, ":"));
	}
	fclose(statfp);
	fputs("</TABLE>",fp);

	fputs("<P><TABLE BORDER=1>",fp);
	fputs("<TR><TH>Empeg Power</TH></TR>",fp);
	statfp = fopen("empeg_power", "r");
	while(fgets(buffer, sizeof(buffer),statfp) != NULL) {
		fprintf(fp, "<TR><TD>%s</TD></TR>\n",
		buffer);
	}
	fclose(statfp);
	fputs("</TABLE>",fp);

	fputs("<P><TABLE BORDER=1>",fp);
	fputs("<TR><TH COLSPAN=2>System Load</TH></TR>",fp);
	fd = open("loadavg", O_RDONLY);
	len = read(fd, buffer, sizeof(buffer)-1);
	close(fd);
	buffer[len] = '\0';
	fprintf(fp, "<TR><TD>1  minute interval</TD><TD>%s</TD></TR>\n",
		strtok(buffer, " "));
	fprintf(fp, "<TR><TD>5  minute interval</TD><TD>%s</TD></TR>\n",
		strtok(NULL, " "));
	fprintf(fp, "<TR><TD>15 minute interval</TD><TD>%s</TD></TR>\n",
		strtok(NULL, " "));
	fputs("</TABLE>",fp);

	fputs("<P><TABLE BORDER=1>",fp);
	fputs("<TR><TH COLSPAN=2>Memory info</TH></TR>",fp);
	fd = open("meminfo", O_RDONLY);
	len = read(fd, buffer, sizeof(buffer)-1);
	close(fd);
	buffer[len] = '\0';
	p = strchr(buffer, '\n');
	p = strtok(p, " ");
	fprintf(fp, "<TR><TD>Memory Total</TD><TD>%s Bytes</TD></TR>\n",
		strtok(NULL, " "));
	fprintf(fp, "<TR><TD>Memory Used</TD><TD>%s Bytes</TD></TR>\n",
		strtok(NULL, " "));
	fprintf(fp, "<TR><TD>Memory Free</TD><TD>%s Bytes</TD></TR>\n",
		strtok(NULL, " "));
	fprintf(fp, "<TR><TD>Shared Memory</TD><TD>%s Bytes</TD></TR>\n",
		strtok(NULL, " "));
	fprintf(fp, "<TR><TD>Buffers</TD><TD>%s Bytes</TD></TR>\n",
		strtok(NULL, " "));
	fprintf(fp, "<TR><TD>Cached</TD><TD>%s Bytes</TD></TR>\n",
		strtok(NULL, "\n"));
	fputs("</TABLE>",fp);

	fputs("<P><TABLE BORDER=1>",fp);
	fputs("<TR><TH COLSPAN=16>ProcessTable</TH></TR>",fp);
	fputs("<TR><TH>Name</TH><TH>State</TH><TH>PID</TH><TH>PPID</TH>",fp);
	fputs("<TH>vsize</TH><TH>rss</TH><TH>priority</TH>",fp);
	fputs("<TH>TTY</TH><TH>minflt</TH><TH>majflt</TH>",fp);
	fputs("<TH>cminflt</TH><TH>cmajflt</TH><TH>utime</TH><TH>stime</TH>",fp);
	fputs("<TH>cutime</TH><TH>cstime</TH></TR>",fp);
	dir = opendir(".");
	while ((ent = readdir(dir)) != NULL) {
		if (isdigit(ent->d_name[0])) {
			get_psentry(fp, atoi((char *)&(ent->d_name[0])));
		}
	}

	fputs("</TABLE></CENTER></H1></FONT>",fp);
	fputs("</BODY></HTML>",fp);
	closedir(dir);
	return 0;
}

int check_auth(struct displayservconn *dsconn, char *src) {
/*
check base64-encoded string for right combination
*/
	char target[256];
	int i;

	i=b64_pton(src,target,sizeof(target)-1);
	target[i]='\0';
	if ( strncmp(target,"ttt:ttt",7) == 0 )
		return 1;
	return 0;
}

int send_auth(FILE *fp) {
/*
send authentication headers
*/
	fputs("HTTP/1.1 401 Authorization Required\n",fp);
	fputs("Server: EmpegCar DisplayServer/1.1 (Linux)\n",fp);
	fputs("WWW-Authenticate: Basic realm=\"displayserver\"\n", fp);
	fputs("Connection: Keep-Alive\n", fp);
	fputs("Content-Type: text/html\n\n",fp);
	fputs("<HTML><TITLE>401 Authorization Required</TITLE>\n", fp);
	fputs("</BODY><H1>Authorization Required</H1></BODY></HTML>\n", fp);
	return 0;
}

int main(int argc, char *argv[]) {

	struct displayserv *ds;
	struct sockaddr_in address;
	int sock, conn, i;
	char *c,*d;
	char mp3fid[20];
	int fd;
	FILE *fp;
	int f=0;
	int sl,serve_type;
	int childpid;
	int auth,doauth=1;
	char buf[2048];
//	struct sched_param *sparam;
	size_t addrLength = sizeof(struct sockaddr_in);
	signal(SIGPIPE,SIG_IGN);
//	mlockall(MCL_CURRENT);
//	sparam=(struct sched_param *)xlloc(sizeof(struct sched_param));
//	sparam->sched_priority=SCHED_RR;
//	sched_setscheduler(0,SCHED_RR,sparam);

#ifdef MEMWATCH
	mwStatistics( 2 );
	mwDoFlush(1);
#endif

	if ( argc > 1 ) {
		doauth=0;
		auth=1;
	}
	f=fork();
	if ( f > 0 )
		exit(0);
	// readconfig;
	ds = (struct displayserv *) malloc( sizeof(struct displayserv));
	ds->lle=NULL;
	ds->fle=NULL;
	ds->logcount=0;
	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
		return 5;

	/* Let the kernel reuse the socket address. This lets us run
	twice in a row, without waiting for the (ip, port) tuple
	to time out. */
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)); 

	address.sin_family = AF_INET;
	address.sin_port = htons(LISTEN_SOCKET);
	memset(&address.sin_addr, 0, sizeof(address.sin_addr));

	if (bind(sock, (struct sockaddr *) &address, sizeof(address)))
		return 6;

	if (listen(sock, 5))
		return 7;


	while ((conn = accept(sock, (struct sockaddr *) &address, 
                          &addrLength)) >= 0) {

		childpid=fork();
		if ( childpid > 0 ) {
			struct displayservconn *dsconn;
			if ( (fp = fdopen( conn, "r+" ) ) == NULL ) {
				fprintf( stderr,"Cannot open stream\n");
				return 8;
			}
			dsconn = (struct displayservconn *)
					malloc( sizeof(struct displayservconn));
			dsconn->ds=ds;
			dsconn->conn=conn;
			dsconn->fp=fp;
			dsconn->vhost=NULL;
			if ( doauth )
				auth=0;
			else
				auth=1;
			while (strlen(fgets(buf, sizeof(buf),fp)) > 2 ) {
#ifdef DEBUG
				fprintf( stderr,"%s",buf);
#endif
				if (strstr(buf,"Host: ")!=NULL) {
					buf[strlen(buf)-2]='\0';
					dsconn->vhost=strdup(&buf[6]);
				}
				if (strstr(buf,"Authorization: Basic ")!=NULL) {
					if ( doauth )
						auth=check_auth(dsconn,
							(char *)&buf[21]);
					else
						auth=1;
				}
				if (strstr(buf,"GET / ")!=NULL)
					serve_type=EMPEG_INDEX;
				if (strstr(buf,"STREAM /empeg_stream")!=NULL) {
					serve_type=EMPEG_STREAM;
					fprintf(stderr,"STREAM1\n");
					auth=1;
					}
				if (strstr(buf,"GET /jscript.html ")!=NULL)
					serve_type=EMPEG_JSCRIPT;
				if (strstr(buf,"GET /empeg.html ")!=NULL)
					serve_type=EMPEG_HTML;
				if (strstr(buf,"GET /empegApplet.class ")!=NULL)
					serve_type=EMPEG_CLASS;
				if (strstr(buf,"GET /fascia.jpg ")!=NULL)
					serve_type=EMPEG_FASCIA;
				if (strstr(buf,"GET /sysinfo.html ")!=NULL)
					serve_type=EMPEG_SYSINFO;
				if (strstr(buf,"GET /log.html ")!=NULL)
					serve_type=EMPEG_LOG;
				if (strstr(buf,"GET /mp3list.html ")!=NULL) {
					strcpy(mp3fid,"100");
					serve_type=EMPEG_MP3PLIST;
				}
				c=strstr(buf,"?playlist=");
				if ( (c!=NULL)&&(strstr(buf,"GET /")!=NULL) ) {
					c+=10;
					strncpy(mp3fid,strtok(c," "),5);
					mp3fid[strlen(mp3fid)-1]='0';
					//printf("debug: pl mp3fid:%s x\n",mp3fid);
					if (strstr(buf,".m3u?playlist=")!=NULL)
						serve_type=EMPEG_M3U;
					else
						serve_type=EMPEG_MP3PLIST;
				}
				c=strstr(buf,".mp3?file=");
				if ( (c!=NULL)&&(strstr(buf,"GET /")!=NULL) ) {
					c+=10;
					//strcpy(mp3fid,"/drive0/fids/");
					strncpy(mp3fid,strtok(c," "),5);
					mp3fid[strlen(mp3fid)-1]='0';
					//printf("debug: fl mp3fid:%s x\n",mp3fid);
					serve_type=EMPEG_MP3FILE;
				}
				c=strstr(buf,"request-delay: ");
				if ( c != NULL ) {
					c+=16;
					sl=strtol(c,NULL,10);
					if ( sl < MINSLEEP || sl > MAXSLEEP )
						sl=DEFSLEEP;
				}
			}
			
		if ( ! auth )
				send_auth(fp);
		
		else switch (serve_type) {
		case EMPEG_INDEX    :	send_index(dsconn);
					break;
		case EMPEG_JSCRIPT  :	send_jscript(fp);
					break;
		case EMPEG_HTML     :	send_html(fp);
					break;
		case EMPEG_CLASS    :	serve_file(dsconn,CLASSFILE);
					break;
		case EMPEG_FASCIA   :	serve_file(dsconn,FASCIAFILE);
					break;
		case EMPEG_STREAM   :	serve_stream(dsconn,sl);
					break;
		case EMPEG_MP3PLIST :	send_mp3list(dsconn,mp3fid);
					break;
		case EMPEG_MP3FILE  :	if ( (chdir("/drive1/fids")) != -1) {
						if ((stat(mp3fid,(struct stat *)
								&buf))!=0)
							chdir("/drive0/fids");
					} else
						chdir("/drive0/fids");
					serve_file(dsconn,mp3fid);
					break;
		case EMPEG_M3U      :	send_m3u(dsconn,mp3fid);
					break;
		case EMPEG_SYSINFO  :	send_sysinfo(dsconn);
					break;
		case EMPEG_LOG      :	send_log(dsconn);
					break;
		}
		fflush(fp);
		shutdown(conn,2);	
		close(conn);
		free(dsconn->vhost);
		free(dsconn);
		free(ds);
		exit(0);
		}
	}
	free(ds);
	if (conn > 0) 
		close(sock);
	return 0;
}


